#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## Trim analysis
# Adrien Crovato
#
# Find the angle of attack to match a specified lift coefficient

import math
import dart.utils as dU
import tbox.utils as tU
from fwk.coloring import ccolors
from dart.scripts.config import Config

class Trim(Config):
    def __init__(self, p):
        # save some parameters for later use
        self.dim = p['Dim']
        self.format = p['Format']
        try:
            self.slice = p['Slice']
            self.tag = p['TagId']
        except:
            self.slice = None
            self.tag = None
        self.mach = p['M_inf']
        self.ClTgt = p['CL']
        self.alpha = p['AoA']*math.pi/180
        self.dCl = p['dCL']
        # basic init
        Config.__init__(self, p)

    def run(self):
        # initialize loop
        it = 0
        while True:
            # define current angle of attack
            print(ccolors.ANSI_BLUE + 'Setting AoA to', self.alpha*180/math.pi, ccolors.ANSI_RESET)
            # update problem
            self.pbl.update(self.alpha)
            # run the solver
            self.tms["solver"].start()
            success = self.solver.run()
            if not success:
                raise RuntimeError('Flow solver diverged!\n')
            self.tms["solver"].stop()
            # update slope
            if it != 0:
                self.dCl = (self.solver.Cl - Cl_) / (self.alpha - alpha_)
            # break or store values and update AoA
            if abs(self.ClTgt - self.solver.Cl) < 0.005 or math.isnan(self.solver.Cl):
                break
            else:
                Cl_ = self.solver.Cl
                alpha_ = self.alpha
                dAlpha = (self.ClTgt - self.solver.Cl) / self.dCl
                self.alpha += dAlpha
                it += 1

        # save results
        self.solver.save(self.mshWriter)
        print('\n')
        # extract Cp
        if self.dim == 2:
            Cp = dU.extract(self.bnd.groups[0].tag.elems, self.solver.Cp)
            tU.write(Cp, "Cp_airfoil.dat", "%1.5e", ", ", "x, y, z, Cp", "")
        elif self.dim == 3 and self.format == 'vtk' and self.slice:
            dU.writeSlices(self.msh.name, self.slice, self.tag)

    def disp(self):
        # display results
        print(ccolors.ANSI_BLUE + 'Trim analysis' + ccolors.ANSI_RESET)
        print("       M    alpha      dCl       Cl       Cd       Cm")
        print("{0:8.2f} {1:8.1f} {2:8.4f} {3:8.4f} {4:8.4f} {5:8.4f}".format(self.mach, self.alpha*180/math.pi, self.dCl, self.solver.Cl, self.solver.Cd, self.solver.Cm))
        print('\n')
        # display timers
        self.tms["total"].stop()
        print(ccolors.ANSI_BLUE + 'CPU statistics' + ccolors.ANSI_RESET)
        print(self.tms)
