#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## Viscous-inviscid coupler (quasi-simultaneous coupling)
# Amaury Bilocq

import numpy as np
from fwk.coloring import ccolors
from dart.viscous.airfoil import Airfoil
from dart.viscous.wake import Wake

class Coupler:
    def __init__(self, _isolver, _vsolver, _boundaryAirfoil, _boundaryWake, _tol, _writer):
        self.isolver =_isolver # inviscid solver
        self.vsolver = _vsolver # viscous solver
        self.group = [Airfoil(_boundaryAirfoil), Wake(_boundaryWake)] # airfoil and wake python objects
        self.tol = _tol # tolerance of the VII
        self.writer = _writer

    def run(self):
        ''' Perform coupling
        '''
        # initialize loop
        it = 0
        converged = 0 # temp
        CdOld = self.vsolver.Cd
        while converged == 0:
            print(ccolors.ANSI_BLUE + 'Iteration: ', it, ccolors.ANSI_RESET)
            # run inviscid solver
            self.isolver.run()
            print('--- Viscous solver parameters ---')
            print('Tolerance (drag count):', self.tol * 1000)
            print('--- Viscous problem definition ---')
            print('Reynolds number:', self.vsolver.Re)
            print('')
            for n in range(0, len(self.group)):
                print('Computing for', self.group[n].name, '...', end = ' ')
                for i in range(0, len(self.group[n].boundary.nodes)):
                    self.group[n].v[i,0] = self.isolver.U[self.group[n].boundary.nodes[i].row][0]
                    self.group[n].v[i,1] = self.isolver.U[self.group[n].boundary.nodes[i].row][1]
                    self.group[n].v[i,2] = self.isolver.U[self.group[n].boundary.nodes[i].row][2]
                    self.group[n].Me[i] = self.isolver.M[self.group[n].boundary.nodes[i].row]
                    self.group[n].rhoe[i] = self.isolver.rho[self.group[n].boundary.nodes[i].row]
                    # run viscous solver
                self.vsolver.run(self.group[n])
                if self.group[n].name == 'airfoil':
                    self.group[n+1].T0 = self.group[n].TEnd[0]+self.group[n].TEnd[1]
                    self.group[n+1].H0 = (self.group[n].HEnd[0]*self.group[n].TEnd[0]+self.group[n].HEnd[1]*self.group[n].TEnd[1])/self.group[n+1].T0
                    self.group[n+1].Ct0 = (self.group[n].CtEnd[0]*self.group[n].TEnd[0]+self.group[n].CtEnd[1]*self.group[n].TEnd[1])/self.group[n+1].T0
                # get blowing velocity from viscous solver and update inviscid solver
                for i in range(0, self.group[n].nE):
                    self.group[n].boundary.setU(i, self.group[n].u[i])
                print('done!')
            print('    Iter           Cd      xtr_top      xtr_bot        Error')
            print('{0:8d} {1:12.5f} {2:12.5f} {3:12.5f} {4:12.5f}'.format(it, self.vsolver.Cd, self.vsolver.xtr[0], self.vsolver.xtr[1], abs(self.vsolver.Cd - CdOld)/self.vsolver.Cd))
            print('')
            # Converged or not
            if abs(self.vsolver.Cd - CdOld)/self.vsolver.Cd < self.tol:
                converged = 1
            else:
                CdOld = self.vsolver.Cd
            it += 1
            self.vsolver.it += 1
        # save results
        self.isolver.save(self.writer)
        self.vsolver.writeFile()
        print('\n')
